#include "pkgi.h"
#include "pkgi_db.h"
#include "pkgi_menu.h"
#include "pkgi_config.h"
#include "pkgi_dialog.h"
#include "pkgi_download.h"
#include "pkgi_utils.h"
#include "pkgi_style.h"

#include <stddef.h>
#include <mini18n.h>
#include <SDL2/SDL.h>

#define content_filter(c)   (c ? 1 << (7 + c) : DbFilterAllContent)

typedef enum  {
    StateError,
    StateRefreshing,
    StateUpdateDone,
    StateMain,
    StateDownload,
    StateTerminate
} State;

static State state;

static uint32_t first_item;
static uint32_t selected_item;

static int search_active;

static char refresh_url[MAX_CONTENT_TYPES][256];

static Config config;
static Config config_temp;

static pkgi_texture background;
static pkgi_input input;

static int font_height;
static int avail_height;
static int bottom_y;

static char search_text[256];
static char error_state[256];

static void reposition(void);
int psp_network_up(void);


static const char* pkgi_get_ok_str(void)
{
    return pkgi_ok_button() == PKGI_BUTTON_X ? PKGI_UTF8_X : PKGI_UTF8_O;
}

static const char* pkgi_get_cancel_str(void)
{
    return pkgi_cancel_button() == PKGI_BUTTON_O ? PKGI_UTF8_O : PKGI_UTF8_X;
}

static void pkgi_refresh_thread(void)
{
    LOG("starting update");

    if (pkgi_menu_result() == MenuResultRefresh)
    {
        pkgi_db_update((char*) &refresh_url, sizeof(refresh_url[0]), error_state, sizeof(error_state));
    }

    if (pkgi_db_reload(error_state, sizeof(error_state)))
    {
        first_item = 0;
        selected_item = 0;
        state = StateUpdateDone;
    }
    else
    {
        state = StateError;
    }
    
    pkgi_thread_exit();
}

static int install(void)
{
    LOG("installing...");
    pkgi_dialog_start_progress(_("Installing..."), _("Please wait..."), -1);

    pkgi_dialog_allow_close(0);
    int ok = pkgi_install(config.install_mode_iso, !config.keep_pkg);
    pkgi_dialog_allow_close(1);

    return ok;
}

static void pkgi_do_download(void)
{
    DbItem* item = pkgi_db_get(selected_item);

    LOG("download thread start");

    pkgi_lock_process();
    if (pkgi_download(item))
    {
        if (install())
        {
            pkgi_dialog_message(item->name, _("Successfully installed"));
            LOG("install succeeded");
        }
        else
        {
            pkgi_dialog_error(_("Installation failed"));
        }
        LOG("download completed!");
    }
    pkgi_unlock_process();

    if (pkgi_dialog_is_cancelled())
    {
        pkgi_dialog_close();
    }

    item->presence = PresenceUnknown;
    state = StateMain;
}

static uint32_t friendly_size(uint64_t size)
{
    if (size > 10ULL * 1000 * 1024 * 1024)
    {
        return (uint32_t)(size / (1024 * 1024 * 1024));
    }
    else if (size > 10 * 1000 * 1024)
    {
        return (uint32_t)(size / (1024 * 1024));
    }
    else if (size > 10 * 1000)
    {
        return (uint32_t)(size / 1024);
    }
    else
    {
        return (uint32_t)size;
    }
}

static const char* friendly_size_str(uint64_t size)
{
    if (size > 10ULL * 1000 * 1024 * 1024)
    {
        return _("GB");
    }
    else if (size > 10 * 1000 * 1024)
    {
        return _("MB");
    }
    else if (size > 10 * 1000)
    {
        return _("KB");
    }
    else
    {
        return _("B");
    }
}

int pkgi_check_free_space(uint64_t size)
{
    uint64_t free = pkgi_get_free_space();

    size *= 2; // we need at least twice the space to install a package
    if (size > free)
    {
        char error[256];
        pkgi_snprintf(error, sizeof(error), _(".pkg requires %u %s free space\n(%s/PKG) only %u %s available"),
            friendly_size(size), friendly_size_str(size),
            pkgi_get_storage_device(),
            friendly_size(free), friendly_size_str(free)
        );

        pkgi_dialog_error(error);
        return 0;
    }

    return 1;
}

static void pkgi_friendly_size(char* text, uint32_t textlen, int64_t size)
{
    if (size <= 0)
    {
        text[0] = 0;
    }
    else if (size < 1000LL)
    {
        pkgi_snprintf(text, textlen, "%u " PKGI_UTF8_B, (uint32_t)size);
    }
    else if (size < 1000LL * 1000)
    {
        pkgi_snprintf(text, textlen, "%.2f " PKGI_UTF8_KB, size / 1024.f);
    }
    else if (size < 1000LL * 1000 * 1000)
    {
        pkgi_snprintf(text, textlen, "%.2f " PKGI_UTF8_MB, size / 1024.f / 1024.f);
    }
    else
    {
        pkgi_snprintf(text, textlen, "%.2f " PKGI_UTF8_GB, size / 1024.f / 1024.f / 1024.f);
    }
}

static const char* content_type_str(ContentType content)
{
    switch (content)
    {
        case 0: return _("All");
        case ContentGame: return _("Games");
        case ContentDLC: return _("DLCs");
        case ContentTheme: return _("Themes");
        case ContentPSX: return _("PSX");
        case ContentDemo: return _("Demos");
        case ContentUpdate: return _("Updates");
        case ContentEmulator: return _("Emulators");
        case ContentApp: return _("Apps");
        case ContentLocal: return _("Local PKGs");
        default: return _("Unknown");
    }
}

static void cb_dialog_exit(int res)
{
    state = StateTerminate;
}

static void cb_dialog_download(int res)
{
    DbItem* item = pkgi_db_get(selected_item);

    item->presence = PresenceMissing;
    pkgi_dialog_start_progress(_("Downloading..."), _("Preparing..."), 0);
    state = StateDownload;
}

static void pkgi_do_main(pkgi_input* input)
{
    int col_titleid = PKGI_MAIN_HMARGIN;
    int col_region = col_titleid + pkgi_text_width("PCSE00000") + PKGI_MAIN_COLUMN_PADDING;
    int col_installed = col_region + pkgi_text_width("USA") + PKGI_MAIN_COLUMN_PADDING;
    int col_name = col_installed + pkgi_text_width(PKGI_UTF8_INSTALLED) + PKGI_MAIN_COLUMN_PADDING;

    uint32_t db_count = pkgi_db_count();
    
    if (input)
    {
        if (input->active & pkgi_cancel_button())
        {
            input->pressed &= ~pkgi_cancel_button();
            pkgi_dialog_ok_cancel("\xE2\x98\x85  PKGi PSP v" PKGI_VERSION "  \xE2\x98\x85", _("Exit to XMB?"), &cb_dialog_exit);
        }

        if (input->active & PKGI_BUTTON_SELECT)
        {
            input->pressed &= ~PKGI_BUTTON_SELECT;
            pkgi_dialog_message("\xE2\x98\x85  PKGi PSP v" PKGI_VERSION "  \xE2\x98\x85",
                                "PlayStation Portable version by Bucanero\n\n"
                                " https://github.com/bucanero/pkgi-psp/");
        }

        if (input->active & PKGI_BUTTON_LEFT)
        {
            config.filter ^= content_filter(config.content);

            if (config.content == 0)
                config.content = MAX_CONTENT_TYPES;

            config.content--;
            config.filter ^= content_filter(config.content);

            pkgi_db_configure(search_active ? search_text : NULL, &config);
            reposition();
            db_count = pkgi_db_count();
        }

        if (input->active & PKGI_BUTTON_RIGHT)
        {
            config.filter ^= content_filter(config.content);

            config.content++;
            if (config.content == MAX_CONTENT_TYPES)
                config.content = 0;

            config.filter ^= content_filter(config.content);

            pkgi_db_configure(search_active ? search_text : NULL, &config);
            reposition();
            db_count = pkgi_db_count();
        }

        if (input->active & PKGI_BUTTON_UP)
        {
            if (selected_item == first_item && first_item > 0)
            {
                first_item--;
                selected_item = first_item;
            }
            else if (selected_item > 0)
            {
                selected_item--;
            }
            else if (selected_item == 0)
            {
                selected_item = db_count - 1;
                uint32_t max_items = avail_height / (font_height + PKGI_MAIN_ROW_PADDING) - 1;
                first_item = db_count > max_items ? db_count - max_items - 1 : 0;
            }
        }

        if (input->active & PKGI_BUTTON_DOWN)
        {
            uint32_t max_items = avail_height / (font_height + PKGI_MAIN_ROW_PADDING) - 1;
            if (selected_item == db_count - 1)
            {
                selected_item = first_item = 0;
            }
            else if (selected_item == first_item + max_items)
            {
                first_item++;
                selected_item++;
            }
            else
            {
                selected_item++;
            }
        }
        
        if (input->active & PKGI_BUTTON_LT)
        {
            uint32_t max_items = avail_height / (font_height + PKGI_MAIN_ROW_PADDING) - 1;
            if (first_item < max_items)
            {
                first_item = 0;
            }
            else
            {
                first_item -= max_items;
            }
            if (selected_item < max_items)
            {
                selected_item = 0;
            }
            else
            {
                selected_item -= max_items;
            }
        }

        if (input->active & PKGI_BUTTON_RT)
        {
            uint32_t max_items = avail_height / (font_height + PKGI_MAIN_ROW_PADDING) - 1;
            if (first_item + max_items < db_count - 1)
            {
                first_item += max_items;
                selected_item += max_items;
                if (selected_item >= db_count)
                {
                    selected_item = db_count - 1;
                }
            }
        }
    }
    
    int y = font_height*3/2 + PKGI_MAIN_HLINE_EXTRA + PKGI_MAIN_VMARGIN;
    int line_height = font_height + PKGI_MAIN_ROW_PADDING;
    for (uint32_t i = first_item; i < db_count; i++)
    {
        DbItem* item = pkgi_db_get(i);

        if (i == selected_item)
        {
            pkgi_draw_fill_rect_z(0, y, PKGI_FONT_Z, PKGI_SCREEN_WIDTH, font_height + PKGI_MAIN_ROW_PADDING - 1, PKGI_COLOR_SELECTED_BACKGROUND);
        }
        uint32_t color = PKGI_COLOR_TEXT;

        char titleid[0x10];
        pkgi_snprintf(titleid, sizeof(titleid), "%.9s", item->content + 7);

        if (item->presence == PresenceUnknown)
        {
            item->presence = pkgi_is_incomplete(item->content) ? PresenceIncomplete : pkgi_is_installed(titleid) ? PresenceInstalled : PresenceMissing;
        }

        char size_str[64];
        pkgi_friendly_size(size_str, sizeof(size_str), item->size);
        int sizew = pkgi_text_width(size_str);

        pkgi_clip_set(0, y, PKGI_SCREEN_WIDTH, line_height);
        pkgi_draw_text(col_titleid, y, color, titleid);
        const char* region;
        switch (pkgi_get_region(item->content))
        {
        case RegionASA: region = "ASA"; break;
        case RegionEUR: region = "EUR"; break;
        case RegionJPN: region = "JPN"; break;
        case RegionUSA: region = "USA"; break;
        default: region = "???"; break;
        }
        pkgi_draw_text(col_region, y, color, region);
        if (item->presence == PresenceIncomplete)
        {
            pkgi_draw_text(col_installed, y, color, PKGI_UTF8_PARTIAL);
        }
        else if (item->presence == PresenceInstalled)
        {
            pkgi_draw_text(col_installed, y, color, PKGI_UTF8_INSTALLED);
        }
        pkgi_draw_text(PKGI_SCREEN_WIDTH - (PKGI_MAIN_SCROLL_WIDTH + PKGI_MAIN_SCROLL_PADDING + PKGI_MAIN_HMARGIN + sizew), y, color, size_str);
        pkgi_clip_remove();

        pkgi_clip_set(col_name, y, PKGI_SCREEN_WIDTH - PKGI_MAIN_SCROLL_WIDTH - PKGI_MAIN_SCROLL_PADDING - PKGI_MAIN_COLUMN_PADDING - sizew - col_name, line_height);
        pkgi_draw_text_ttf(0, 0, PKGI_FONT_Z, color, item->name);
        pkgi_clip_remove();

        y += font_height + PKGI_MAIN_ROW_PADDING;
        if (y > PKGI_SCREEN_HEIGHT - (font_height + PKGI_MAIN_HLINE_EXTRA*6 + PKGI_MAIN_VMARGIN))
        {
            break;
        }
        else if (y + font_height > PKGI_SCREEN_HEIGHT - (font_height + PKGI_MAIN_HLINE_EXTRA))
        {
            line_height = (PKGI_SCREEN_HEIGHT - (font_height + PKGI_MAIN_HLINE_EXTRA)) - (y + 1);
            if (line_height < PKGI_MAIN_ROW_PADDING)
            {
                break;
            }
        }
    }

    if (db_count == 0)
    {
        const char* text = _("No items!");

        int w = pkgi_text_width(text);
        pkgi_draw_text((PKGI_SCREEN_WIDTH - w) / 2, PKGI_SCREEN_HEIGHT / 2, PKGI_COLOR_TEXT, text);
    }

    // scroll-bar
    if (db_count != 0)
    {
        uint32_t max_items = (avail_height + font_height + PKGI_MAIN_ROW_PADDING - 1) / (font_height + PKGI_MAIN_ROW_PADDING) - 1;
        if (max_items < db_count)
        {
            uint32_t min_height = PKGI_MAIN_SCROLL_MIN_HEIGHT;
            uint32_t height = max_items * avail_height / db_count;
            uint32_t start = first_item * (avail_height - (height < min_height ? min_height : 0)) / db_count;
            height = max32(height, min_height);
            pkgi_draw_fill_rect_z(PKGI_SCREEN_WIDTH - (PKGI_MAIN_HMARGIN + PKGI_MAIN_SCROLL_WIDTH), font_height + PKGI_MAIN_HLINE_EXTRA + PKGI_MAIN_VMARGIN + start + 2, PKGI_FONT_Z, PKGI_MAIN_SCROLL_WIDTH, height, PKGI_COLOR_SCROLL_BAR);
        }
    }

    if (input && (input->pressed & pkgi_ok_button()) && db_count)
    {
        input->pressed &= ~pkgi_ok_button();

        DbItem* item = pkgi_db_get(selected_item);

        if (!pkgi_check_free_space(item->size))
        {
            LOG("[%.9s] %s - no free space", item->content + 7, item->name);
        }
        else if (item->presence == PresenceInstalled)
        {
            LOG("[%.9s] %s - already installed", item->content + 7, item->name);
            pkgi_dialog_ok_cancel(item->name, _("Item already installed, download again?"), &cb_dialog_download);
        }
        else if (item->presence == PresenceIncomplete || (item->presence == PresenceMissing))
        {
            LOG("[%.9s] %s - starting to install", item->content + 7, item->name);
            pkgi_dialog_start_progress(_("Downloading..."), _("Preparing..."), 0);
            state = StateDownload;
        }
    }
    else if (input && (input->pressed & PKGI_BUTTON_T))
    {
        input->pressed &= ~PKGI_BUTTON_T;

        config_temp = config;

        pkgi_menu_start(search_active, &config);
    }
    else if (input && (input->active & PKGI_BUTTON_S) && db_count)
    {
        input->pressed &= ~PKGI_BUTTON_S;

        DbItem* item = pkgi_db_get(selected_item);

        pkgi_dialog_details(item, content_type_str(item->type));
    }
}

static void pkgi_do_refresh(void)
{
    char text[256];

    uint32_t updated;
    uint32_t total;
    pkgi_db_get_update_status(&updated, &total);

    if (total == 0)
    {
        pkgi_snprintf(text, sizeof(text), "%s... %.2f %s", _("Refreshing"), (uint32_t)updated / 1024.f, _("KB"));
    }
    else
    {
        pkgi_snprintf(text, sizeof(text), "%s... %u%%", _("Refreshing"), updated * 100U / total);
    }

    int w = pkgi_text_width(text);
    pkgi_draw_text((PKGI_SCREEN_WIDTH - w) / 2, PKGI_SCREEN_HEIGHT / 2, PKGI_COLOR_TEXT, text);
}

static void pkgi_do_head(void)
{
    char size[64];
    char title[256];
    uint32_t count = pkgi_db_count();
    uint32_t total = pkgi_db_total();

    if (count == total)
    {
        pkgi_snprintf(size, sizeof(size), "%u", count);
    }
    else
    {
        pkgi_snprintf(size, sizeof(size), "%u (%u)", count, total);
    }
    pkgi_snprintf(title, sizeof(title), "PKGi v%s - %s: %s", PKGI_VERSION, content_type_str(config.content), size);
    pkgi_draw_text(PKGI_MAIN_HMARGIN, PKGI_MAIN_VMARGIN, PKGI_COLOR_TEXT_HEAD, title);

    pkgi_draw_fill_rect(0, font_height + PKGI_MAIN_VMARGIN, PKGI_SCREEN_WIDTH, PKGI_MAIN_HLINE_HEIGHT, PKGI_COLOR_HLINE);

    pkgi_friendly_size(size, sizeof(size), pkgi_get_free_space());
    pkgi_snprintf(title, sizeof(title), "%s: %s", _("Free"), size);

    if (search_active)
    {
        pkgi_snprintf(title, sizeof(title), ">> %s <<", search_text);
    }

    int rightw = pkgi_text_width(title);
    pkgi_draw_text(PKGI_SCREEN_WIDTH - PKGI_MAIN_HLINE_EXTRA - (rightw + PKGI_MAIN_HMARGIN), PKGI_MAIN_VMARGIN, PKGI_COLOR_TEXT_HEAD, title);
}

static void pkgi_do_tail(void)
{
    char text[256];
    int batt, status;

    pkgi_draw_fill_rect_z(0, bottom_y - font_height/2, PKGI_FONT_Z, PKGI_SCREEN_WIDTH, PKGI_MAIN_HLINE_HEIGHT, PKGI_COLOR_HLINE);

    batt = pkgi_get_battery_charge(&status);
    pkgi_snprintf(text, sizeof(text), "%c %d%%", status + 0xE9, batt);
    pkgi_draw_text_z(PKGI_MAIN_TEXT_PADDING, bottom_y, PKGI_FONT_Z, PKGI_COLOR_TEXT_TAIL, text);

    if (pkgi_menu_is_open())
    {
        pkgi_snprintf(text, sizeof(text), "%s %s  " PKGI_UTF8_T " %s  %s %s", pkgi_get_ok_str(), _("Select"), _("Close"), pkgi_get_cancel_str(), _("Cancel"));
    }
    else
    {
        pkgi_snprintf(text, sizeof(text), "%s %s  " PKGI_UTF8_T " %s  " PKGI_UTF8_S " %s  %s %s", pkgi_get_ok_str(), _("Download"), _("Menu"), _("Details"), pkgi_get_cancel_str(), _("Exit"));
    }

    pkgi_draw_text_z((PKGI_SCREEN_WIDTH - pkgi_text_width(text)) / 2, bottom_y, PKGI_FONT_Z, PKGI_COLOR_TEXT_TAIL, text);
}

static void pkgi_do_error(void)
{
    pkgi_draw_text((PKGI_SCREEN_WIDTH - pkgi_text_width(error_state)) / 2, PKGI_SCREEN_HEIGHT / 2, PKGI_COLOR_TEXT_ERROR, error_state);
}

static void reposition(void)
{
    uint32_t count = pkgi_db_count();
    if (first_item + selected_item < count)
    {
        return;
    }

    uint32_t max_items = (avail_height + font_height + PKGI_MAIN_ROW_PADDING - 1) / (font_height + PKGI_MAIN_ROW_PADDING) - 1;
    if (count > max_items)
    {
        uint32_t delta = selected_item - first_item;
        first_item = count - max_items;
        selected_item = first_item + delta;
    }
    else
    {
        first_item = 0;
        selected_item = 0;
    }
}

static void pkgi_update_check(void)
{
    char *buffer;
    uint32_t size;

    LOG("checking latest pkgi version at %s", PKGI_UPDATE_URL);

    buffer = pkgi_http_download_buffer(PKGI_UPDATE_URL, &size);

    if (!buffer)
    {
        LOG("no update data available");
        pkgi_thread_exit();
    }

    LOG("received %u bytes", size);

    static const char find[] = "\"name\":\"PKGi PSP v";
    const char* start = strstr(buffer, find);
    if (!start)
    {
        LOG("no name found");
        goto end_update;
    }

    LOG("found name");
    start += sizeof(find) - 1;

    char* end = strchr(start, '"');
    if (!end)
    {
        LOG("no end of name found");
        goto end_update;
    }
    *end = 0;
    LOG("latest version is %s", start);

    if (pkgi_stricmp(PKGI_VERSION, start) == 0)
    {
        LOG("no need to update");
        goto end_update;
    }

    start = strstr(end+1, "\"browser_download_url\":\"");
    if (!start)
        goto end_update;

    start += 24;
    end = strchr(start, '"');
    if (!end)
    {
        LOG("no download URL found");
        goto end_update;
    }

    *end = 0;
    LOG("download URL is %s", start);

    pkgi_dialog_message("New PKGi Version Available!", "Update PKGi PSP to the latest version:\nhttps://github.com/bucanero/pkgi-psp/");
/*
    DbItem update_item = {
        .content = "UP0001-NP00PKGI3_00-0000000000000000",
        .name    = "PKGi PSP Update",
        .url     = start,
    };

    pkgi_dialog_start_progress(update_item.name, _("Preparing..."), 0);
    
    if (pkgi_download(&update_item) && install())
    {
        pkgi_dialog_message(update_item.name, _("Successfully downloaded PKGi PSP update"));
        LOG("update downloaded!");
    }
*/

end_update:
    free(buffer);
}

static void pkgi_load_language(const char* lang)
{
    char path[256];

    pkgi_snprintf(path, sizeof(path), PKGI_APP_FOLDER "/LANG/%s.po", lang);
    LOG("Loading language file (%s)...", path);
    mini18n_set_locale(path);
}

void progress_screen_refresh(void)
{
    pkgi_update(&input);

    pkgi_draw_background(background);
    pkgi_do_head();
    pkgi_do_dialog(&input);
    pkgi_do_tail();

    pkgi_swap();
}


int main(int argc, char* argv[])
{
    SDL_Init(0);
    pkgi_start();

    pkgi_load_config(&config, (char*) &refresh_url, sizeof(refresh_url[0]));
    pkgi_load_language(config.language);
    pkgi_is_psp_go(config.storage);
    pkgi_dialog_init();
    
    font_height = pkgi_text_height("M");
    avail_height = PKGI_SCREEN_HEIGHT - 2 * (font_height + PKGI_MAIN_HLINE_EXTRA*2 + PKGI_MAIN_VMARGIN);
    bottom_y = PKGI_SCREEN_HEIGHT - (PKGI_MAIN_VMARGIN + font_height);

    state = StateRefreshing;
    pkgi_start_thread("refresh_thread", &pkgi_refresh_thread);

    background = pkgi_load_image_buffer(background, png);

    memset(&input, 0, sizeof(pkgi_input));
    while (pkgi_update(&input) && (state != StateTerminate))
    {
        pkgi_draw_background(background);

        if (state == StateUpdateDone)
        {
            if (psp_network_up() && config.version_check)
            {
                pkgi_update_check();
            }

            pkgi_db_configure(NULL, &config);
            state = StateMain;
        }

        pkgi_do_head();
        switch (state)
        {
        case StateError:
            pkgi_do_error();
            // leave the menu open if there's no database and we have URLs available
            if (!pkgi_menu_is_open() && config.allow_refresh && psp_network_up())
            {
                config_temp = config;
                pkgi_menu_start(search_active, &config);
            }            
            break;

        case StateRefreshing:
            pkgi_do_refresh();
            break;

        case StateDownload:
            pkgi_do_download();
            break;

        case StateMain:
            pkgi_do_main(pkgi_dialog_is_open() || pkgi_menu_is_open() ? NULL : &input);
            break;

        default:
            // never happens, just to shut up the compiler
            break;
        }

        pkgi_do_tail();

        if (pkgi_dialog_is_open())
        {
            pkgi_do_dialog(&input);

            if (pkgi_dialog_is_cancelled())
            {
                pkgi_dialog_close();
            }
        }

        if (pkgi_dialog_input_update())
        {
            search_active = 1;
            pkgi_dialog_input_get_text(search_text, sizeof(search_text));
            pkgi_db_configure(search_text, &config);
            reposition();
        }

        if (pkgi_menu_is_open())
        {
            if (pkgi_do_menu(&input))
            {
                Config new_config;
                pkgi_menu_get(&new_config);
                if (config_temp.sort != new_config.sort ||
                    config_temp.order != new_config.order ||
                    config_temp.filter != new_config.filter)
                {
                    config_temp = new_config;
                    pkgi_db_configure(search_active ? search_text : NULL, &config_temp);
                    reposition();
                }
            }
            else
            {
                MenuResult mres = pkgi_menu_result();
                if (mres == MenuResultSearch)
                {
                    pkgi_dialog_input_text(_("Search"), search_text);
                }
                else if (mres == MenuResultSearchClear)
                {
                    search_active = 0;
                    search_text[0] = 0;
                    pkgi_db_configure(NULL, &config);
                }
                else if (mres == MenuResultCancel)
                {
                    if (config_temp.sort != config.sort || config_temp.order != config.order || config_temp.filter != config.filter)
                    {
                        pkgi_db_configure(search_active ? search_text : NULL, &config);
                        reposition();
                    }
                }
                else if (mres == MenuResultAccept)
                {
                    pkgi_menu_get(&config);
                    pkgi_save_config(&config, (char*) &refresh_url, sizeof(refresh_url[0]));
                }
                else if (mres == MenuResultRefresh)
                {
                    state = StateRefreshing;
                    pkgi_start_thread("refresh_thread", &pkgi_refresh_thread);
                }
            }
        }

        pkgi_swap();
    }

    LOG("finished");
    mini18n_close();
    pkgi_free_texture(background);
    pkgi_end();
	return 0;
}
